@@ -1,31 +1,37 @@
 /****************************************************************************
 **
-** Copyright (C) 2014 Digia Plc and/or its subsidiary(-ies).
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtQuick module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL21$
+** $QT_BEGIN_LICENSE:LGPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia. For licensing terms and
-** conditions see http://qt.digia.com/licensing. For further information
-** use the contact form at http://qt.digia.com/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file. Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
-** In addition, as a special exception, Digia gives you certain additional
-** rights. These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -38,6 +44,7 @@
 #include <QtCore/QtMath>
 
 #include <QtGui/QOpenGLContext>
+#include <QtGui/QOpenGLFunctions>
 #include <QtGui/QGuiApplication>
 #include <QtGui/QScreen>
 #include <QtGui/QSurface>
@@ -54,21 +61,13 @@ QT_BEGIN_NAMESPACE
 #define GL_BGRA 0x80E1
 #endif
 
+int qt_sg_envInt(const char *name, int defaultValue);
 
 static QElapsedTimer qsg_renderer_timer;
 
 namespace QSGAtlasTexture
 {
 
-static int qsg_envInt(const char *name, int defaultValue)
-{
-    QByteArray content = qgetenv(name);
-
-    bool ok = false;
-    int value = content.toInt(&ok);
-    return ok ? value : defaultValue;
-}
-
 Manager::Manager()
     : m_atlas(0)
 {
@@ -79,8 +78,8 @@ Manager::Manager()
     int max;
     gl->functions()->glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max);
 
-    int w = qMin(max, qsg_envInt("QSG_ATLAS_WIDTH", qMax(512U, qNextPowerOfTwo(surfaceSize.width() - 1))));
-    int h = qMin(max, qsg_envInt("QSG_ATLAS_HEIGHT", qMax(512U, qNextPowerOfTwo(surfaceSize.height() - 1))));
+    int w = qMin(max, qt_sg_envInt("QSG_ATLAS_WIDTH", qMax(512U, qNextPowerOfTwo(surfaceSize.width() - 1))));
+    int h = qMin(max, qt_sg_envInt("QSG_ATLAS_HEIGHT", qMax(512U, qNextPowerOfTwo(surfaceSize.height() - 1))));
 
     if (surface->surfaceClass() == QSurface::Window) {
         QWindow *window = static_cast<QWindow *>(surface);
@@ -91,7 +90,7 @@ Manager::Manager()
         }
     }
 
-    m_atlas_size_limit = qsg_envInt("QSG_ATLAS_SIZE_LIMIT", qMax(w, h) / 2);
+    m_atlas_size_limit = qt_sg_envInt("QSG_ATLAS_SIZE_LIMIT", qMax(w, h) / 2);
     m_atlas_size = QSize(w, h);
 
     qCDebug(QSG_LOG_INFO, "texture atlas dimensions: %dx%d", w, h);
@@ -112,13 +111,16 @@ void Manager::invalidate()
     }
 }
 
-QSGTexture *Manager::create(const QImage &image)
+QSGTexture *Manager::create(const QImage &image, bool hasAlphaChannel)
 {
-    QSGTexture *t = 0;
+    Texture *t = 0;
     if (image.width() < m_atlas_size_limit && image.height() < m_atlas_size_limit) {
         if (!m_atlas)
             m_atlas = new Atlas(m_atlas_size);
+        // t may be null for atlas allocation failure
         t = m_atlas->create(image);
+        if (t && !hasAlphaChannel && t->hasAlphaChannel())
+            t->setHasAlphaChannel(false);
     }
     return t;
 }
@@ -127,6 +129,7 @@ Atlas::Atlas(const QSize &size)
     : m_allocator(size)
     , m_texture_id(0)
     , m_size(size)
+    , m_atlas_transient_image_threshold(0)
     , m_allocated(false)
 {
 
@@ -137,28 +140,35 @@ Atlas::Atlas(const QSize &size)
     if (QOpenGLContext::currentContext()->isOpenGLES()) {
 #endif
 
-#if defined(Q_OS_ANDROID) && !defined(Q_OS_ANDROID_NO_SDK)
+#if defined(Q_OS_ANDROID) && !defined(Q_OS_ANDROID_EMBEDDED)
     QString *deviceName =
             static_cast<QString *>(QGuiApplication::platformNativeInterface()->nativeResourceForIntegration("AndroidDeviceName"));
     static bool wrongfullyReportsBgra8888Support = deviceName != 0
-                                                    && (deviceName->compare(QStringLiteral("samsung SM-T211"), Qt::CaseInsensitive) == 0
-                                                        || deviceName->compare(QStringLiteral("samsung SM-T210"), Qt::CaseInsensitive) == 0
-                                                        || deviceName->compare(QStringLiteral("samsung SM-T215"), Qt::CaseInsensitive) == 0);
+                                                    && (deviceName->compare(QLatin1String("samsung SM-T211"), Qt::CaseInsensitive) == 0
+                                                        || deviceName->compare(QLatin1String("samsung SM-T210"), Qt::CaseInsensitive) == 0
+                                                        || deviceName->compare(QLatin1String("samsung SM-T215"), Qt::CaseInsensitive) == 0);
 #else
     static bool wrongfullyReportsBgra8888Support = false;
+    // The Raspberry Pi (both 1 and 2) GPU refuses framebuffers with BGRA color attachments.
+    const GLubyte *renderer = QOpenGLContext::currentContext()->functions()->glGetString(GL_RENDERER);
+    if (renderer && strstr((const char *) renderer, "VideoCore IV"))
+        wrongfullyReportsBgra8888Support = true;
 #endif // ANDROID
 
+    if (qEnvironmentVariableIsSet("QSG_ATLAS_NO_BGRA_WORKAROUNDS"))
+        wrongfullyReportsBgra8888Support = false;
+
     const char *ext = (const char *) QOpenGLContext::currentContext()->functions()->glGetString(GL_EXTENSIONS);
-    if (!wrongfullyReportsBgra8888Support
+    if (ext && !wrongfullyReportsBgra8888Support
             && (strstr(ext, "GL_EXT_bgra")
                 || strstr(ext, "GL_EXT_texture_format_BGRA8888")
                 || strstr(ext, "GL_IMG_texture_format_BGRA8888"))) {
         m_internalFormat = m_externalFormat = GL_BGRA;
-#ifdef Q_OS_IOS
-    } else if (strstr(ext, "GL_APPLE_texture_format_BGRA8888")) {
+#if defined(Q_OS_DARWIN) && !defined(Q_OS_OSX)
+    } else if (ext && strstr(ext, "GL_APPLE_texture_format_BGRA8888")) {
         m_internalFormat = GL_RGBA;
         m_externalFormat = GL_BGRA;
-#endif // IOS
+#endif // IOS || TVOS
     } else {
         m_internalFormat = m_externalFormat = GL_RGBA;
     }
@@ -169,6 +179,11 @@ Atlas::Atlas(const QSize &size)
 
     m_use_bgra_fallback = qEnvironmentVariableIsSet("QSG_ATLAS_USE_BGRA_FALLBACK");
     m_debug_overlay = qEnvironmentVariableIsSet("QSG_ATLAS_OVERLAY");
+
+    // images smaller than this will retain their QImage.
+    // by default no images are retained (favoring memory)
+    // set to a very large value to retain all images (allowing quick removal from the atlas)
+    m_atlas_transient_image_threshold = qt_sg_envInt("QSG_ATLAS_TRANSIENT_IMAGE_THRESHOLD", 0);
 }
 
 Atlas::~Atlas()
@@ -377,28 +392,40 @@ void Atlas::bind(QSGTexture::Filtering filtering)
     // Upload all pending images..
     for (int i=0; i<m_pending_uploads.size(); ++i) {
 
-        bool profileFrames = QSG_LOG_TIME_TEXTURE().isDebugEnabled() ||
-                QQuickProfiler::profilingSceneGraph();
+        bool profileFrames = QSG_LOG_TIME_TEXTURE().isDebugEnabled();
         if (profileFrames)
             qsg_renderer_timer.start();
 
+        Q_QUICK_SG_PROFILE_START(QQuickProfiler::SceneGraphTexturePrepare);
+
+        // Skip bind, convert, swizzle; they're irrelevant
+        Q_QUICK_SG_PROFILE_SKIP(QQuickProfiler::SceneGraphTexturePrepare,
+                                QQuickProfiler::SceneGraphTexturePrepareStart, 3);
+
+        Texture *t = m_pending_uploads.at(i);
         if (m_externalFormat == GL_BGRA &&
                 !m_use_bgra_fallback) {
-            uploadBgra(m_pending_uploads.at(i));
+            uploadBgra(t);
         } else {
-            upload(m_pending_uploads.at(i));
+            upload(t);
         }
+        const QSize textureSize = t->textureSize();
+        if (textureSize.width() > m_atlas_transient_image_threshold ||
+                textureSize.height() > m_atlas_transient_image_threshold)
+            t->releaseImage();
 
         qCDebug(QSG_LOG_TIME_TEXTURE).nospace() << "atlastexture uploaded in: " << qsg_renderer_timer.elapsed()
-                                           << "ms (" << m_pending_uploads.at(i)->image().width() << "x"
-                                           << m_pending_uploads.at(i)->image().height() << ")";
+                                           << "ms (" << t->textureSize().width() << "x"
+                                           << t->textureSize().height() << ")";
 
-        Q_QUICK_SG_PROFILE(QQuickProfiler::SceneGraphTexturePrepare, (
-                0,  // bind (not relevant)
-                0,  // convert (not relevant)
-                0,  // swizzle (not relevant)
-                qsg_renderer_timer.nsecsElapsed(), // (upload all of the above)
-                0)); // mipmap (not used ever...)
+        Q_QUICK_SG_PROFILE_RECORD(QQuickProfiler::SceneGraphTexturePrepare,
+                                  QQuickProfiler::SceneGraphTexturePrepareUpload);
+
+        // Skip mipmap; unused
+        Q_QUICK_SG_PROFILE_SKIP(QQuickProfiler::SceneGraphTexturePrepare,
+                                QQuickProfiler::SceneGraphTexturePrepareUpload, 1);
+        Q_QUICK_SG_PROFILE_REPORT(QQuickProfiler::SceneGraphTexturePrepare,
+                                  QQuickProfiler::SceneGraphTexturePrepareMipmap);
     }
 
     GLenum f = filtering == QSGTexture::Nearest ? GL_NEAREST : GL_LINEAR;
@@ -423,15 +450,15 @@ Texture::Texture(Atlas *atlas, const QRect &textureRect, const QImage &image)
     , m_image(image)
     , m_atlas(atlas)
     , m_nonatlas_texture(0)
+    , m_has_alpha(image.hasAlphaChannel())
 {
-    m_allocated_rect_without_padding = m_allocated_rect.adjusted(1, 1, -1, -1);
     float w = atlas->size().width();
     float h = atlas->size().height();
-
-    m_texture_coords_rect = QRectF(m_allocated_rect_without_padding.x() / w,
-                                   m_allocated_rect_without_padding.y() / h,
-                                   m_allocated_rect_without_padding.width() / w,
-                                   m_allocated_rect_without_padding.height() / h);
+    QRect nopad = atlasSubRectWithoutPadding();
+    m_texture_coords_rect = QRectF(nopad.x() / w,
+                                   nopad.y() / h,
+                                   nopad.width() / w,
+                                   nopad.height() / h);
 }
 
 Texture::~Texture()
@@ -448,15 +475,63 @@ void Texture::bind()
 
 QSGTexture *Texture::removedFromAtlas() const
 {
-    if (!m_nonatlas_texture) {
+    if (m_nonatlas_texture) {
+        m_nonatlas_texture->setMipmapFiltering(mipmapFiltering());
+        m_nonatlas_texture->setFiltering(filtering());
+        return m_nonatlas_texture;
+    }
+
+    if (!m_image.isNull()) {
         m_nonatlas_texture = new QSGPlainTexture();
         m_nonatlas_texture->setImage(m_image);
         m_nonatlas_texture->setFiltering(filtering());
-        m_nonatlas_texture->setMipmapFiltering(mipmapFiltering());
+
+    } else {
+        QOpenGLFunctions *f = QOpenGLContext::currentContext()->functions();
+        // bind the atlas texture as an fbo and extract the texture..
+
+        // First extract the currently bound fbo so we can restore it later.
+        GLint currentFbo;
+        f->glGetIntegerv(GL_FRAMEBUFFER_BINDING, &currentFbo);
+
+        // Create an FBO and bind the atlas texture into it.
+        GLuint fbo;
+        f->glGenFramebuffers(1, &fbo);
+        f->glBindFramebuffer(GL_FRAMEBUFFER, fbo);
+        f->glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, m_atlas->textureId(), 0);
+
+        // Create the target texture, QSGPlainTexture below will deal with the texparams, so we don't
+        // need to worry about those here.
+        GLuint texture;
+        f->glGenTextures(1, &texture);
+        f->glBindTexture(GL_TEXTURE_2D, texture);
+        QRect r = atlasSubRectWithoutPadding();
+        // and copy atlas into our texture.
+        while (f->glGetError() != GL_NO_ERROR) ;
+        f->glCopyTexImage2D(GL_TEXTURE_2D, 0, m_atlas->internalFormat(), r.x(), r.y(), r.width(), r.height(), 0);
+        // BGRA may have been rejected by some GLES implementations
+        if (f->glGetError() != GL_NO_ERROR)
+            f->glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, r.x(), r.y(), r.width(), r.height(), 0);
+
+        m_nonatlas_texture = new QSGPlainTexture();
+        m_nonatlas_texture->setTextureId(texture);
+        m_nonatlas_texture->setOwnsTexture(true);
+        m_nonatlas_texture->setHasAlphaChannel(m_has_alpha);
+        m_nonatlas_texture->setTextureSize(r.size());
+
+        // cleanup: unbind our atlas from the fbo, rebind the old default and delete the fbo.
+        f->glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, 0, 0);
+        f->glBindFramebuffer(GL_FRAMEBUFFER, (GLuint) currentFbo);
+        f->glDeleteFramebuffers(1, &fbo);
     }
+
+    m_nonatlas_texture->setMipmapFiltering(mipmapFiltering());
+    m_nonatlas_texture->setFiltering(filtering());
     return m_nonatlas_texture;
 }
 
 }
 
 QT_END_NAMESPACE
+
+#include "moc_qsgatlastexture_p.cpp"
